#include "Lektor2.h"
#include "audio.h"
#include <analogWrite.h>
#include "spd.h"
#include <driver/adc.h>
#include <esp_adc_cal.h>
#include <freertos/ringbuf.h>
#include <ESP32Encoder.h>
#include <Bounce2.h>
#include <Preferences.h>


#include "EMeter.h"
#include "calib.h"
#include "ohmset.h"



ESP32Encoder encoder;
Bounce encKey;

struct WiFiPrefs wiprefs;
Preferences preferences;

extern void readVolts(bool force = false);

SpeechdClient spd("miernik:pomiar",nullptr,0,1);
Lektor2 spiker;
AudioOut audio(12);
char spkbuf[32],oldspk[32];
char dspbuf[10];
bool beeping=false;
RingbufHandle_t SpeechRB;
volatile bool audioRunning;

esp_adc_cal_characteristics_t hars1, hars2;

bool lowOhm=false;

int speechMode=SPEMD_KLATT;
TaskHandle_t speechHandle;

int presetmv=5000;
int presetstep;


#define OHM_TOO_LARGE -2
#define OHM_TOO_SMALL -1
#define OHM_UNKNOWN -3

#define lowLed 10

void setLed(int led)
{
    static int ledValue=0;
    if (ledValue >= 0 && ledValue != led) {
        analogWrite(PIN_LED,ledValue = led);
    }
}


void setOhmRange(int n)
{
    if (!n) n=-1;
    if (n < 0) {
        adc1_config_channel_atten(ADC1_OHM,ADC_ATTEN_DB_0);
        n = -n;
        lowOhm = true;
    }
    else {
        adc1_config_channel_atten(ADC1_OHM,ADC_ATTEN_DB_11);
        lowOhm=false;
    }

    digitalWrite(PIN_OHM_0, (n-1)&1);
    digitalWrite(PIN_OHM_1, (n-1)&2);
    digitalWrite(PIN_OHM_2, !((n-1)&4));
}

void paszcza(int a)
{
    static int pa[]={0,10,80,255};
    a&=3;
    setLed(pa[a]);
}


struct spikerCmd {
    uint8_t cmd;
    char text[255];
    };
    
void speechThread(void *v)
{
    size_t itemsize;
    spiker.setMouthCallback(paszcza);
    spiker.setAudioModule(&audio);
    struct spikerCmd *rbc;
    for (;;) {
        audioRunning = false;
        rbc = (struct spikerCmd *) xRingbufferReceive(SpeechRB, &itemsize, portMAX_DELAY);
        if (rbc) {
            audioRunning=true;
            if (rbc->cmd) spiker.say(rbc->text);
            else {
                int freq = (rbc->text[0] & 255) | ((rbc->text[1] << 8) & 0xff00);
                spiker.beep(freq);
            }
            vRingbufferReturnItem(SpeechRB, (void *)rbc);
        }
    }
}


void relaySet(int8_t a, int8_t b, bool force=false)
{
    if (a) digitalWrite(PIN_RL1A, 1); else digitalWrite(PIN_RL1B, 1);
    if (b) digitalWrite(PIN_RL2A, 1); else digitalWrite(PIN_RL2B, 1);
    delay(200);
    digitalWrite(PIN_RL1A,0);
    digitalWrite(PIN_RL1B,0);
    digitalWrite(PIN_RL2A,0);
    digitalWrite(PIN_RL2B,0);
    
}

#define relaySetVoltLo relaySet(1,1)
#define relaySetVoltHi relaySet(0,1)
#define relaySetVolt relaySetVoltHi
#define relaySetOhm relaySet(0,0)
#define relaySetDiode relaySet(0,0)
#define relaySetShort relaySet(1,0)
#define relaySetInit relaySet(1,0,1)


void storeWiprefs(void)
{
    preferences.putBytes("nets",(void *)&wiprefs, sizeof(wiprefs));
}


void setup()
{
    pinMode(PIN_OHM_0, OUTPUT);
    pinMode(PIN_OHM_1, OUTPUT);
    pinMode(PIN_OHM_2, OUTPUT);
    setOhmRange(4);
    pinMode(PIN_RL1A, OUTPUT);digitalWrite(PIN_RL1A,0);
    pinMode(PIN_RL1B, OUTPUT);digitalWrite(PIN_RL1B,0);
    pinMode(PIN_RL2A, OUTPUT);digitalWrite(PIN_RL2A,0);
    pinMode(PIN_RL2B, OUTPUT);digitalWrite(PIN_RL2B,0);

    encKey.attach(PIN_KEY_ENCODER,INPUT_PULLUP);
    encKey.interval(50);
    
    ESP32Encoder::useInternalWeakPullResistors=UP;
    encoder.attachHalfQuad(PIN_ENCODER_A, PIN_ENCODER_B);

    adc1_config_width(ADC_WIDTH_BIT_12);
    adc1_config_channel_atten(ADC1_ACC,ADC_ATTEN_DB_11);
    adc1_config_channel_atten(ADC1_HIGH,ADC_ATTEN_DB_11);
    adc1_config_channel_atten(ADC1_OHM,ADC_ATTEN_DB_11);
    adc1_config_channel_atten(ADC1_VOLT,ADC_ATTEN_DB_11);
    esp_adc_cal_characterize(ADC_UNIT_1, ADC_ATTEN_DB_11, ADC_WIDTH_BIT_12, 1090, &hars1);
    esp_adc_cal_characterize(ADC_UNIT_1, ADC_ATTEN_DB_0, ADC_WIDTH_BIT_12, 1090, &hars2);
    Serial.begin(115200);
    initDisplay();
    preferences.begin("miernik");
    int z=preferences.getBytes("nets",(void *)&wiprefs, sizeof(wiprefs));
    if (z != sizeof(wiprefs)) {
        preferences.remove("nets");
        memset((void *)&wiprefs, 0, sizeof(wiprefs));
    }
    if (wiprefs.speechd_addr) spd.setHost(wiprefs.speechd_addr);
    wiprefs.app_burst = constrain(wiprefs.app_burst,2,30);
    wiprefs.raw_burst = constrain(wiprefs.raw_burst,2,30);
    speechMode = preferences.getUChar("spmod",SPEMD_KLATT);
    presetmv = preferences.getShort("precise",5000);
    speechMode = constrain(speechMode, 0, SPEMD_MAX-1);
    
    readVolts(true);
    encKey.update();
    encKey.update();
    if (!encKey.read()){
        delay(100);
        encKey.update();
        if (!encKey.read()) AP_Start();
    }
    if (speechMode == SPEMD_APP || speechMode == SPEMD_RAW || speechMode == SPEMD_SPEECHD) {
        WiFi_Start();
    }
    SpeechRB = xRingbufferCreateNoSplit(256,4);
    xTaskCreatePinnedToCore(speechThread,"speech",10000,NULL,3,&speechHandle,0);
    if (wiprefs.module)spd.setModule(wiprefs.module);
    if (wiprefs.voice)spd.setVoice(wiprefs.voice);
    spd.setRate(wiprefs.rate);
    spd.setLanguage("pl");
    relaySetVoltHi;
    setOhmRange(4);
    setLed(lowLed);
    mainMode = MMOD_SELECT;
    setDisplayBig(DPMOD_SELECT);
    selShow=0;
    doDisplay();
}


int anaCompa(const void *a, const void *b)
{
    return *(short*)a - *(short*)b;
}

int analogVoltage(adc_channel_t channel)
{
#define MEASURES 63
    short table[MEASURES];
    int i;
    for (i=0;i<MEASURES;i++) {
        uint32_t volt;
        esp_adc_cal_get_voltage(channel,(lowOhm && (channel == ADC_OHM))? &hars2: &hars1,&volt);
        table[i] = volt;
    }
    qsort(table,MEASURES,sizeof(short),anaCompa);
    return table[MEASURES/2];
}

void stopAudio(bool speech = false)
{
    audio.stop();
    if (speech && speechMode == SPEMD_SPEECHD) spd.stop();
    while(audioRunning) delay(10);
}

void beepbeep(unsigned int freq=440)
{
    uint8_t rbc[3];
    rbc[0]=0;
    rbc[1]=freq & 255;
    rbc[2]=freq >>8;
    stopAudio();
    if (xRingbufferSend(SpeechRB, rbc, sizeof(rbc), 0) == pdTRUE) {
        beeping=1;
    }
}

void say(const char *text)
{
    stopAudio(true);
    struct spikerCmd rbc;
    rbc.cmd = 1;
    strcpy(rbc.text, text);
    xRingbufferSend(SpeechRB, &rbc, sizeof(rbc), 0);

}


int reada(void)
{
    char buf[32];
    if (!Serial.available()) return -1;
    int n=Serial.readBytesUntil('\n',buf,31);
    buf[n]=0;
    return strtol(buf,nullptr,10);
}

int getEncoder(void)
{
    static int64_t lastEncoder=0;
    int delta;
    int64_t enc = encoder.getCount();
    delta = enc - lastEncoder;
    lastEncoder = enc;
    return delta;
    
}


int mV_Vcc, mV_LiIon;
static uint32_t lastVolt_a=0, lastVolt_h=0;
void readVolts(bool force)
{
    uint32_t ms=millis();
    
    if (force || ms - lastVolt_a >= 1000UL) {
        lastVolt_a=ms;
        mV_Vcc = (mV_Vcc + VOLT33) / 2;
    }
    if (force || ms - lastVolt_h >= 1000UL) {
        lastVolt_h=ms;
        mV_LiIon = VOLT74;
    }
}

void readVcc(void)
{
   lastVolt_a=millis();
   mV_Vcc = (mV_Vcc + VOLT33) / 2;
}


static float u2ohm(double U, double R, double u)
{
    return u * R/(U-u);
}

static int readOhm(void)
{
    readVcc();
    int i,j,u;
    for (i=0;i<OHMSETCNT;i++) {
        setOhmRange(OHMSETCNT-i-1);
        delay(1);
        analogVoltage(ADC_OHM);
        delay(1);
        u = (analogVoltage(ADC_OHM) * mV_Vcc) / 3300;
        if (u > ohmsets[i].umax) {
            if (i) return OHM_UNKNOWN;
            return OHM_TOO_LARGE;
        }
        for (j=0;j<ohmsets[i].nranges;j++) {
            if (u >= ohmsets[i].ranges[j].lora) {
                return (int)u2ohm(ohmsets[i].ranges[j].U,ohmsets[i].ranges[j].R,(double)u);
            }
        }
    }
    return OHM_TOO_SMALL;
}

static const int szereg[]={10, 11, 12, 13, 15, 16, 18, 20, 22, 24, 27, 30, 33, 36, 39, 43, 47, 51, 56, 62, 68, 75, 82, 91 };

int szereguj(int v, int step)
{
    int t = 110;
    int mpx = 1;
    while (v >= t * mpx) {
        mpx *= 10;
    }
    for (int i=step; i<=23; i+=2*step) if (v < mpx * szereg[i]) return mpx * szereg[i-step];
    return 100 * mpx;
}


#define Volt0 (VOLT0 + (mV_Vcc - 3300)/2)
#define VMPX (1000.0/(VOLT10 - VOLT0))
#define VMPX24 (100.0/(VOLT24 - VOLT0))

static int realvolt(bool hi)
{
    readVcc();
    int rv=analogVoltage(ADC_VOLT);
    int mv= rv - Volt0;
    int sgn = 1;
    if (mv < 0) {
        mv = -mv;
        sgn = -1;
    }
    if (mv < 8) mv = 0; // próg działania coby nie jęczał przy zerze
    if (hi) {
        int decivolts = ((mv * VMPX24)+0.5);
        return sgn * decivolts * 100;
    }
    int centivolts = ((mv * VMPX)+0.5);
    centivolts = centivolts / 5;
    return sgn * centivolts * 50;
}

static uint8_t meast;
static uint32_t lastMeas;

static bool meas(void)
{
    if (audioRunning) {
        meast = 1;
        return false;
    }
    if (meast == 1) {
        meast = 2;
        lastMeas = millis();
        return false;
    }
    uint32_t delta = 300;
    if (speechMode == SPEMD_RAW) delta = 100 * wiprefs.raw_burst;
    else if (speechMode == SPEMD_APP) delta = 100 * wiprefs.app_burst;
    if (meast == 2 && millis() - lastMeas >= delta) {
        meast = 0;
    }
    return meast == 0;
}


void csay(const char *c, int v)
{
    static const char *rawnames[]={
        "?","R20","R10","R","?","D","V10","V24"};
    
    char rawbuf[32];
    if (!speechMode) return;
    
    if (speechMode != SPEMD_RAW && !strcmp(c,oldspk)) return;
    doDisplay();
    strcpy(oldspk,c);
    switch(speechMode) {
        case SPEMD_KLATT:
        setLed(0);
        say(c);
        break;

        case SPEMD_SPEECHD:
        spd.say(c);
        break;

        case SPEMD_RAW:
        sprintf(rawbuf,"%s %d\n", rawnames[mainMode],v);
        sendUDPBuf(wiprefs.raw_addr, wiprefs.raw_port,rawbuf);
        break;

        case SPEMD_APP:
        strcpy(rawbuf,c);strcat(rawbuf,"\n");
        sendUDPBuf(wiprefs.app_addr, wiprefs.app_port,rawbuf);
        break;
    }
}
void setCalibRelay()
{
    switch (thisCalib) {
        case CLB_VHI:
        case CLB_UAC:
        case CLB_VCC:
        setOhmRange(4);
        relaySetVoltHi;break;

        case CLB_VLO:
        setOhmRange(4);
        relaySetVoltLo;break;

        case CLB_SHO:
        setOhmRange(-1);
        relaySetShort;break;
        
        case CLB_DIO:
        setOhmRange(2);
        relaySetDiode;break;

        default:
        setOhmRange(thisCalib - CLB_R0);
        relaySetOhm;break;
    }
}

static const char *calibs="HLUVDP01234";

void pre2step(void)
{
    int pl=presetmv, ph=0;
    if (pl > 10000) {
        ph =presetmv - pl;
        pl = 10000;
    }
    presetstep = pl/50 + ph/100;
}

void step2pre(void)
{
    int pl = presetstep, ph = 0;
    if (pl > 200) {
        ph = pl - 200;
        pl = 200;
    }
    presetmv=pl * 50 + ph * 100;
}

#define MAXPRESET 260
uint32_t lastPrecBeep;
void loop()
{
    
    int freq;
    if (speechMode == SPEMD_SPEECHD) {
        spd.loop();
    }
    readVolts();
    encKey.update();
    int t=getEncoder();
    int resistor, voltage;
    if (mainMode == MMOD_SELECT) {
        if (t) {
            stopAudio(true);
            beeping=0;
            selShow += t;
            selShow = constrain(selShow,0,MMOD_MAX-2);
            doDisplay();
            return;
        }
        if (encKey.fell()) {
            mainMode = selShow+1;
            oldspk[0]=0;
            meast=0;
            switch (mainMode) {
                case MMOD_PRECIZE:
                setOhmRange(4);
                if (presetmv > 8000) {
                    relaySetVoltHi;
                }
                else {
                    relaySetVoltLo;
                }
                lastPrecBeep=millis();
                break;
                
                case MMOD_PRESET:
                setOhmRange(4);
                relaySetVoltHi;
                pre2step();
                formatVolt(presetmv, dspbuf, spkbuf);
                setDisplayBig(DPMOD_VALUE, dspbuf);
                doDisplay();
                return;
                
                case MMOD_VOLTAGE:
                setOhmRange(4);
                relaySetVoltLo;
                break;

                case MMOD_HIVOLTAGE:
                setOhmRange(4);
                relaySetVoltHi;
                break;

                case MMOD_DIODE:
                relaySetOhm;
                setOhmRange(2);
                break;

                case MMOD_SHORT:
                relaySetShort;
                setOhmRange(-1);
                break;

                case MMOD_OHM_STILL:
                case MMOD_OHM_TYP6:
                case MMOD_OHM_TYP12:
                relaySetOhm;
                setOhmRange(4);
                break;

                case MMOD_PREFS:
                setDisplayBig(DPMOD_SPEECH);
                initSpkDispModes();
                doDisplay();
                return;

                case MMOD_WIFI:
                setDisplayBig(DPMOD_WIFI);
                doDisplay();
                return;

                case MMOD_CALIB:
                setDisplayBig(DPMOD_CALIB,"");
                thisCalib=0;
                oldCalib=-1;
                setCalibRelay();
                break;
                
                default:
                mainMode = MMOD_SELECT;
                selShow=0;
                return;
            }
            setDisplayBig(DPMOD_NONE);
            doDisplay();
            return;
                            
        }
        delay(10);
        doDisplay();
        return;
    }
    if (encKey.fell()) {
        stopAudio(true);
        beeping=0;
        if (mainMode == MMOD_PRECIZE) preferences.putShort("precise",presetmv);
        if (speechMode == SPEMD_SPEECHD) spd.stop();
        selShow = mainMode - 1;
        mainMode = MMOD_SELECT;
        if (dpMOD == DPMOD_SPEECH) {
            if (speechMode != spkdisplay[thisSpkMode].spkmode) {
                speechMode = spkdisplay[thisSpkMode].spkmode;
                preferences.putUChar("spmod",speechMode);
            }
            if (speechMode != SPEMD_SPEECHD) spd.stop();
            if (speechMode == SPEMD_SPEECHD ||
                    speechMode == SPEMD_APP ||
                    speechMode == SPEMD_RAW) {
                WiFi_Start();
            }
            else {
                WiFi_Stop();
            }
        }
        relaySetVoltHi;
        setDisplayBig(DPMOD_SELECT);
        doDisplay();
        return;
    }
    switch (mainMode) {
        case MMOD_PRESET:
        if (t) {
            presetstep += t;
            presetstep=constrain(presetstep,0,MAXPRESET);
            step2pre();
                
            formatVolt(presetmv, dspbuf, spkbuf);
            setDisplayBig(DPMOD_VALUE, dspbuf);
        }
        break;

        case MMOD_PRECIZE:
        voltage = realvolt(presetmv > 8000);
        formatVolt(voltage, dspbuf, spkbuf);
        setDisplayBig(DPMOD_VALUE, dspbuf);
        voltage = abs(voltage);
        if (voltage >= presetmv) {
            freq= presetmv+(voltage - presetmv) * 5;
        }
        else {
            freq = presetmv+(voltage - presetmv) * 2;
        }
        freq = (440 * freq)/presetmv;
        freq = constrain(freq,220,880);
        
        printf("Beepvolt = %d\n",freq);
        if (!beeping) {
            beepbeep(freq);
        }
        else audio.setbeep(freq);
        break;
        
        
        case MMOD_CALIB:
        if (t) {
            thisCalib += t;
            thisCalib = constrain(thisCalib,0,CLB_MAX-1);
            if (thisCalib != oldCalib) setCalibRelay();
        }
        else {
            if (!meas()) break;
            voltage = analogVoltage(
                (thisCalib == CLB_UAC)? ADC_ACC:
                (thisCalib == CLB_VCC)? ADC_HIGH:
                (thisCalib <= CLB_VLO) ? ADC_VOLT:
                ADC_OHM);
            sprintf(dspbuf,"%c %4d", calibs[thisCalib],voltage);
            setDisplayBig(DPMOD_CALIB,dspbuf);
            meast=1;
        }
        break;

        
        case MMOD_PREFS:
        thisSpkMode += t;
        thisSpkMode = constrain(thisSpkMode, 0, maxSpkMode);
        break;
        
        case MMOD_OHM_TYP6:
        case MMOD_OHM_TYP12:
        case MMOD_OHM_STILL:
        if (!meas()) break;
        resistor = readOhm();
        if (resistor == -1) {
            setDisplayBig(DPMOD_ZWARCIE);
            csay("zwarcie",resistor);
        }
        else if (resistor < 0) {
            setDisplayBig(DPMOD_NONE);
            oldspk[0]=0;
        }
        else {
            if (mainMode == MMOD_OHM_TYP6) resistor = szereguj(resistor,2);
            else if (mainMode == MMOD_OHM_TYP12) resistor = szereguj(resistor,1);
            formatOhm(resistor,dspbuf,spkbuf);
            setDisplayBig(DPMOD_VALUE, dspbuf);
            csay(spkbuf, resistor);
        
        }
        meast=1;
        break;
        
        case MMOD_VOLTAGE:
        case MMOD_HIVOLTAGE:
        if (!meas()) break;

        voltage = realvolt(mainMode == MMOD_HIVOLTAGE);
        formatVolt(voltage, dspbuf, spkbuf);
        setDisplayBig(DPMOD_VALUE, dspbuf);
        meast=1;
        csay(spkbuf, voltage);
        break;

        case MMOD_DIODE:
        if (!meas()) break;
        voltage = analogVoltage(ADC_OHM);
        if (voltage < 900) {
            setOhmRange(-2);
            analogVoltage(ADC_OHM);
            delay(10);
            voltage = analogVoltage(ADC_OHM);
            setOhmRange(2);
            analogVoltage(ADC_OHM);
        }
        if (voltage < 90) {
            setDisplayBig(DPMOD_ZWARCIE);
            csay("zwarcie",0);
        }
        else if (voltage > 3000) {
            setDisplayBig(DPMOD_NONE);
            oldspk[0]=0;
        }
        else {
            formatVolt(voltage,dspbuf, spkbuf);
            setDisplayBig(DPMOD_VALUE, dspbuf);
            csay(spkbuf, voltage);
        }
        meast=1;
        break;

        case MMOD_SHORT:
        voltage = analogVoltage(ADC_OHM);
        if (voltage < MINSHORTBEEP && !beeping) {
            beepbeep();
        }
        else if (voltage >=MAXSHORTBEEP && beeping) {
            stopAudio();
            beeping=false;
        }
        setLed(beeping ? 255 : lowLed);
        break;

    }
    if (mainMode == MMOD_PRECIZE) {
        if (millis() - lastPrecBeep >= 200UL) {
            lastPrecBeep = millis();
            doDisplay();
        }
    }
    else doDisplay();
        
    if (!audioRunning) {
        int mt = millis() & 511;
        int mv = (mV_LiIon - 6400) / 2;
        mv = constrain(mv, 50,512);
        setLed((mt <= mv) ? lowLed:0);
    }
}
