#include <Arduino.h>
#include <WiFi.h>
#include <WiFiClient.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include "EMeter.h"
#include <ESPmDNS.h>

const char *ap_ssid="5G Vaccine Gateway";
const char *ap_pass="billgates";

static AsyncWebServer server(80);

bool apActive = false;

static void notFound(AsyncWebServerRequest *request) {
    request->send(404, "text/plain", "Not found");
}

static const char *make_message(const char *fmt, ...)
{
   int size = 0;
   char *p = NULL;
   va_list ap;

   va_start(ap, fmt);
   size = vsnprintf(p, size, fmt, ap);
   va_end(ap);

   if (size < 0)
       return NULL;

   size++;
   p = (char *)malloc(size);
   if (p == NULL)
       return NULL;

   va_start(ap, fmt);
   size = vsnprintf(p, size, fmt, ap);
   va_end(ap);

   if (size < 0) {
       free(p);
       return NULL;
   }

   return p;
}


const char *ifmt=R"D00P(
<!DOCTYPE html>
<html>
<head>
<meta name="viewport" content="width=device-width, initial-scale=1.0">
<meta http-equiv="content-type" content="text/html; charset=utf-8">
<title>Ustawienia WiFi miernika</title>
</head>
<body>
%s
<form method="POST" action="/preset">
<fieldset>
<legend>Ustawienia sieci</legend>
<table>
<tbody>
<tr><td><label for="ssid_i">SSID</label></td><td><input type="text" name="ssid" id="ssid_i" value="%s"></td></tr>
<tr><td><label for="pass_i">Hasło</label></td><td><input type="text" name="pass" id="pass_i" value="%s"></td></tr>
<tr><td><label for="ipadr_i">IP</label></td><td><input type="text" name="ipadr" id="ipadr_i" value="%d.%d.%d.%d"></td></tr>
<tr><td><label for="gway_i">Gateway</label></td><td><input type="text" name="gway" id="gway_i" value="%d.%d.%d.%d"></td></tr>
<tr><td><label for="mask_i">Netmask</label></td><td><input type="text" name="mask" id="mask_i" value="%d.%d.%d.%d"></td></tr>
<tr><td><label for="sport_i">Speechd: port</label></td><td><input type="number" step="1" min="0" max="65535" name="sport" id="sport_i" value="%d"></td></tr>
<tr><td><label for="spip_i">Speechd: IP</label></td><td><input type="text" name="spip" id="spip_i" value="%d.%d.%d.%d"></td></tr>
<tr><td><label for="smod_i">Speechd: module</label></td><td><input type="text" name="smod" id="smod_i" value="%s"></td></tr>
<tr><td><label for="svox_i">Speechd: voice</label></td><td><input type="text" name="svox" id="svox_i" value="%s"></td></tr>
<tr><td><label for="srat_i">Speechd: rate</label></td><td><input type="number" min="-100" max="100" step="1" name="srat" id="srat_i" value="%d"></td></tr>
<tr><td><label for="aport_i">App: port</label></td><td><input type="number" step="1" min="0" max="65535" name="aport" id="aport_i" value="%d"></td></tr>
<tr><td><label for="apip_i">App: IP</label></td><td><input type="text" name="apip" id="spip_i" value="%d.%d.%d.%d"></td></tr>
<tr><td><label for="abrst_i">App: burst</label></td><td><input type="text" name="abrst" id="abrst_i" value="%d">&nbsp;*&nbsp;100&nbsp;msec</td></tr>
<tr><td><label for="rport_i">Raw: port</label></td><td><input type="number" step="1" min="0" max="65535" name="rport" id="rport_i" value="%d"></td></tr>
<tr><td><label for="rpip_i">Raw: IP</label></td><td><input type="text" name="rpip" id="rpip_i" value="%d.%d.%d.%d"></td></tr>
<tr><td><label for="rbrst_i">Raw: burst</label></td><td><input type="number" min="0" max="30" name="rbrst" id="rbrst_i" value="%d">&nbsp;*&nbsp;100&nbsp;msec</td></tr>

</tbody>
</table>
<hr>
<input type="submit" value="Zapisz">
</fieldset>
</form>
</body>
</html>
)D00P";


static const char *htmlescape(const char *in, char *out)
{
    char znak;
    char *rc=out;
    while ((znak = *in++)) {
        switch(znak) {
            case '<':
            strcpy(out,"&lt;");out += 4;break;
            case '>':
            strcpy(out,"&gt;");out += 4;break;
            case '&':
            strcpy(out,"&amp;");out += 5;break;
            case '"':
            strcpy(out,"&quot;");out += 6;break;
            default:
            *out++=znak;
        }
    }
    *out=0;
    return rc;
}


static void startPage(AsyncWebServerRequest *request)
{
    char buf1[256], buf2[256];
    const char *s="";
    if (request->hasArg("zapis")) {
        String st=request->arg("zapis");
        s= (st.c_str()[0] == '1') ? "<b>Zmiany zostały zapisane</b><br>" : "<b style='color:red'>Błędne dane</b><br>";
    }
    const char *buffer=make_message(ifmt,
        s,
        htmlescape(wiprefs.ssid,buf1),
        htmlescape(wiprefs.password,buf2),
        int2ip(wiprefs.ipadr),
        int2ip(wiprefs.gateway),
        int2ip(wiprefs.mask),
        wiprefs.speechd_port,
        int2ip(wiprefs.speechd_addr),
        wiprefs.module,
        wiprefs.voice,
        wiprefs.rate,
        wiprefs.app_port,
        int2ip(wiprefs.app_addr),
        wiprefs.app_burst,
        wiprefs.raw_port,
        int2ip(wiprefs.raw_addr),
        wiprefs.raw_burst
        );
    if (buffer) {
        request->send(200,"text/html",buffer);
        free((void *)buffer);
    }
    else {
        request->send(500,"text/plain","Internal server error");
    }
}

static bool getStrField(char *whereto, AsyncWebServerRequest *request, const char *field, int maxlen)
{
    if (!request->hasArg(field)) {
        printf("Brak <%s>\n", field);
        return false;
    }
    String s=request->arg(field);
    if (s.length() > maxlen) return false;
    strcpy(whereto, s.c_str());
    return true;
}

static bool getIntField(int *whereto, AsyncWebServerRequest *request, const char *field)
{
    if (!request->hasArg(field)) {
        printf("Brak <%s>\n", field);
        return false;
    }
    String s=request->arg(field);
    const char *c=s.c_str();
    while (*c && isspace(*c)) c++;
    if (!*c) {
        *whereto=0;
        return true;
    }
    *whereto = strtol(c,(char **)&c,10);
    while (*c && isspace(*c)) c++;
    if (*c) return false;
    return true;
}

static bool getIPField(uint32_t *whereto, AsyncWebServerRequest *request, const char *field)
{
    if (!request->hasArg(field)) {
        printf("Brak <%s>\n", field);
        return false;
    }
    String s=request->arg(field);
    IPAddress ip;
    ip.fromString(s);
    *whereto = (uint32_t)ip;
    return true;
}
 

static bool makePrefs(AsyncWebServerRequest *request)
{
    struct WiFiPrefs lprefs=wiprefs;
    if (!getStrField(lprefs.ssid,request,"ssid",31)) return false;
    if (!getStrField(lprefs.password,request,"pass",31)) return false;
    if (!getStrField(lprefs.module,request,"smod",31)) return false;
    if (!getStrField(lprefs.voice,request,"svox",31)) return false;
    if (!getIntField(&lprefs.speechd_port,request,"sport")) return false;
    if (!getIntField(&lprefs.rate,request,"srat")) return false;
    if (!getIPField(&lprefs.ipadr,request,"ipadr")) return false;
    if (!getIPField(&lprefs.gateway,request,"gway")) return false;
    if (!getIPField(&lprefs.mask,request,"mask")) return false;
    if (!getIPField(&lprefs.speechd_addr,request,"spip")) return false;
    if (!getIPField(&lprefs.raw_addr,request,"rpip")) return false;
    if (!getIPField(&lprefs.app_addr,request,"rpip")) return false;
    if (!getIntField(&lprefs.app_port,request,"aport")) return false;
    if (!getIntField(&lprefs.raw_port,request,"rport")) return false;
    if (!getIntField(&lprefs.app_burst,request,"abrst")) return false;
    if (!getIntField(&lprefs.raw_burst,request,"rbrst")) return false;
    wiprefs = lprefs;
    return true;
}

static void setPreset(AsyncWebServerRequest *request)
{
    if (!makePrefs(request)) {
        request->redirect("/?zapis=0");
    }
    else {
        storeWiprefs();
        request->redirect("/?zapis=1");
    }
}

void AP_Start(void)
{
    if (apActive) return;
    if (wiprefs.ssid[0]) {
        WiFi.mode(WIFI_AP_STA);
        WiFi.begin(wiprefs.ssid, wiprefs.password);
        if (wiprefs.ipadr) {
            WiFi.config(IPAddress(wiprefs.ipadr), IPAddress(wiprefs.gateway), IPAddress(wiprefs.mask));
        }
        MDNS.begin("miernik");
    }
    else {
        WiFi.mode(WIFI_AP);
    }
    WiFi.softAP(ap_ssid,ap_pass);
    Serial.println(WiFi.softAPIP()); 
    server.onNotFound(notFound);
    server.on("/",HTTP_GET,startPage);
    server.on("/preset",HTTP_POST,setPreset);
    server.begin();
    apActive=true;
}

void AP_Stop(void)
{
    if (!apActive) return;
    apActive = false;
    server.end();
#ifdef WIFI_DEMO
    WiFi.disconnect();
#else
    WiFi.softAPdisconnect();
#endif
}

WiFiUDP sudp;
static bool ustart=false;
bool wstart = false;
void sendUDPBuf(uint32_t addr, int port, const char *str)
{
    if (WiFi.status() != WL_CONNECTED) return;
    if (!ustart) {
        ustart=true;
        printf("LIP %s\n",WiFi.localIP().toString().c_str());
        sudp.begin(WiFi.localIP(), 3459);
    }
    sudp.beginPacket(IPAddress(addr), port);
    printf("Sending to %s:%d %s",IPAddress(addr).toString().c_str(),port,str);
    sudp.write((const uint8_t *)str, strlen(str));
    sudp.endPacket();
}

void WiFi_Start(void)
{
    if (apActive) return;
    if (wstart) return;
    if (!wiprefs.ssid[0]) return;
    WiFi.mode(WIFI_STA);
    if (wiprefs.password[0]) {
        WiFi.begin(wiprefs.ssid, wiprefs.password);
    }
    else {
        WiFi.begin(wiprefs.ssid);
    }
    if (wiprefs.ipadr) {
        WiFi.config(IPAddress(wiprefs.ipadr), IPAddress(wiprefs.gateway), IPAddress(wiprefs.mask));
    }
    wstart = true;
}

    
void WiFi_Stop(void)
{
    if (apActive) return;
    if (wstart) WiFi.disconnect();
    wstart = false;
    ustart = false;
}


