
#include <Wire.h>
#include <WiFi.h>
#include <Adafruit_GFX.h>
#include "Adafruit_SH1106.h"
#include "font9x15.h"
#include "font8x13.h"
#include "bigfont.h"

#include "EMeter.h"


void i2c_reset(void)
{
    digitalWrite(I2C_RST, 0);
    delay(10);
    digitalWrite(I2C_RST, 1);
}



Adafruit_SH1106 display(I2C_RST);

void drawBigChar(Adafruit_GFX &disp, int x, int y, uint8_t c, int color=WHITE)
{
    if (c <=32 || c >= bigfont_max) return;
    int n = (c-33) * 120;
    disp.drawBitmap(x,y,bigfont+n,17,40,color);
    
}

void drawBigText(Adafruit_GFX &disp,int x, int y, const char *str, int color=WHITE, int bgcolor=-1, bool compact=false, int len=0)
{
    if (!len) len=strlen(str);
    if (bgcolor >= 0) {
        disp.fillRect(x,y,(compact?18:21)*strlen(str),40,bgcolor);
    }
    for (;*str && len>0;str++, len--) {
        drawBigChar(disp, x,y,(uint8_t )*str, color);
        x += compact?18:21;
    }
}

void drawChar9x15(int x, int y,uint8_t c, int color=WHITE)
{
    int n = (c & 128) ? 96 * 15 : 0;
    n += 15 * ((c & 127) - 32);
    display.drawBitmap(x, y, font9x15 + n, 8,15, color);
}

void drawText9x15(int x, int y, const char *c, bool bold=false, int color=WHITE)
{
    if (x < 0) {
        int w = (bold ? 9 : 8) * strlen(c);
        x = 64 - w/2;
    }
    for (;*c;c++) {
        drawChar9x15(x,y,(uint8_t)*c, color);
        if (bold) {
            x++;
            drawChar9x15(x,y,(uint8_t)*c,color);
        }
        x+=8;
    }
}

void drawChar8x13(int x, int y,uint8_t c, int color=WHITE)
{
    int n = (c & 128) ? 96 * 13 : 0;
    n += 13 * ((c & 127) - 32);
    display.drawBitmap(x, y, font8x13 + n, 8,13, color);
}

void drawText8x13(int x, int y, const char *c, bool bold=false, int color=WHITE)
{
    if (x < 0) {
        int w = (bold ?  8: 7) * strlen(c);
        x = 64 - w/2;
    }
    for (;*c;c++) {
        drawChar8x13(x,y,(uint8_t)*c, color);
        if (bold) {
            x++;
            drawChar8x13(x,y,(uint8_t)*c,color);
        }
        x+=7;
    }
}


void initDisplay(void)
{
    pinMode(I2C_RST, OUTPUT);
    digitalWrite(I2C_RST, 1);
    delay(100);
    Wire.begin(I2C_SDA, I2C_SCL,400000);
    delay(100);
    i2c_reset();
    display.begin(SH1106_SWITCHCAPVCC, 0x3C);
    display.setTextColor(WHITE, BLACK);
    display.setTextSize(1);
    display.clearDisplay();
    drawBigText(display,11,15,"HELLO");
    display.display();
    delay(1000);

}

static int lmv_Vcc=-1, lmv_LiIon=-1;
int dpMOD, oldDPMOD=-1;
char dispval[10], olddispval[10]="";

static const char *modenames[] = {
    "Wybierz","Rezystor 20%", "Rezystor 10%", "Omomierz","Przej\266cie","Dioda",
    "Napi\352cie 10V", "Napi\352cie 24V","Nap. strojenia","Strojenie","Mowa","WiFi","Kalibracja"
    };

static const char *spkmodes[]={
    "Wy\263\261cz","Wbudowana","WiFi Speechd","WiFi Speaker","WiFi Raw"};

struct spkdisps  spkdisplay[SPEMD_MAX];
int thisSpkMode, oldSpkMode,maxSpkMode;
void initSpkDispModes(void)
{
    int i,n;
    thisSpkMode = 0;
    for (i=n=0;i<SPEMD_MAX;i++) {
        if (i<2 || (wiprefs.ssid[0] && (
                (i == SPEMD_SPEECHD && wiprefs.speechd_addr) ||
                (i == SPEMD_APP && wiprefs.app_addr && wiprefs.app_port) ||
                (i == SPEMD_RAW && wiprefs.raw_addr && wiprefs.raw_port)))) {
            spkdisplay[n].spkmode = i;
            spkdisplay[n].modename = spkmodes[i];
            if (speechMode == i) thisSpkMode = n;
            n++;
        }
    }
    maxSpkMode = n-1;
    oldSpkMode = -1;
}

void setDisplayBig(int mode, const char *str)
{
    dpMOD = mode;
    if (str) strcpy(dispval, str);
    else dispval[0]=0;
}


int8_t mainMode, oldMainMode=-1;
int8_t selShow, oldSelShow=-1;
int8_t thisCalib, oldCalib;
uint8_t wifidisp, oldwifidisp=255;

static const uint8_t PROGMEM wifiIcon[]={
    0x7f,0x00,
    0x80,0x80,
    0x3e,0,
    0x41,0,
    0x1c,0,
    0,0,
    0x08,0
    };

static const uint8_t PROGMEM apIcon[] = {
    0x41,0,
    0x80,0x80,
    0xa2,0x80,
    0xaa,0x80,
    0xaa,0x80,
    0x88,0x80,
    0x41,0
    };
void doDisplay(void)
{
    wifidisp = apActive ? 1 : 0;
    if (WiFi.status() == WL_CONNECTED) wifidisp |= 2;
    bool dpm=
        lmv_Vcc != mV_Vcc/10 || lmv_LiIon != mV_LiIon/10 ||
        oldDPMOD != dpMOD ||
        oldMainMode != mainMode ||
        wifidisp != oldwifidisp ||
        (dpMOD == DPMOD_SELECT && selShow != oldSelShow) ||
        ((dpMOD == DPMOD_VALUE || dpMOD == DPMOD_CALIB) && strcmp(dispval, olddispval)) ||
        (dpMOD == DPMOD_SPEECH && thisSpkMode != oldSpkMode) ||
        (dpMOD == DPMOD_CALIB && thisCalib != oldCalib);
    char buf[32];
    if (!dpm) return;
    display.clearDisplay();
    display.setCursor(0,57);
    sprintf(buf,"Vin=%d.%02d Vcc=%d.%02d",mV_LiIon/1000, (mV_LiIon % 1000) / 10,
        mV_Vcc/1000, (mV_Vcc % 1000) / 10);
    display.print(buf);
    if (wifidisp & 1) {
        display.drawBitmap(108,57,apIcon,9,7,WHITE);
    }
    if (wifidisp & 2) {
        display.drawBitmap(119,57,wifiIcon,9,7,WHITE);
    }
    
    char *c;
    int i,bg,y;
    switch(dpMOD) {
        case DPMOD_WIFI:
        y=15;
        if (!wifidisp) {
            drawText9x15(-1,20,"WY\243\241CZONY");
        }
        else {
            if (apActive) {
                sprintf(buf,"AP:  %s",WiFi.softAPIP().toString().c_str());
                display.setCursor(0,y);display.print(buf); y+= 10;
            }
            if (wifidisp & 2) {
                sprintf(buf,"STA: %s", WiFi.localIP().toString().c_str());
                display.setCursor(0,y);display.print(buf); y+= 10;
            }
        }
        break;
        
        case DPMOD_CALIB:
        drawBigText(display,0,15,dispval);
        break;
        
        case DPMOD_SPEECH:
        bg = thisSpkMode - 1;
        if (bg > maxSpkMode - 2) bg = maxSpkMode - 2;
        if (bg < 0) bg= 0;
        for (i=0;i<3 && i <= maxSpkMode;i++) {
            int col = WHITE;
            if (i+bg == thisSpkMode) {
                display.fillRect(0,13*i+14,128,14,WHITE);
                col = BLACK;
                
            }
            drawText8x13(14,14+13*i,spkdisplay[i+bg].modename,i+bg == thisSpkMode,col);
        }
        break;

        
        case DPMOD_SELECT:
        bg = selShow - 1;
        if (bg < 0) bg = 0;
        if (bg > MMOD_MAX-4) bg = MMOD_MAX-4;
        for (i=0;i<3;i++) {
            int col = WHITE;
            if (i+bg == selShow) {
                display.fillRect(0,13*i+14,128,14,WHITE);
                col = BLACK;
                
            }
            drawText8x13(14,14+13*i,modenames[i+bg+1],i+bg == selShow,col);
        }
        break;
        

        
        case DPMOD_PRZERWA:
        drawBigText(display,1,15,"PRZERWA",WHITE,-1,true);
        break;

        case DPMOD_ZWARCIE:
        drawBigText(display,1,15,"ZWARCIE",WHITE,-1,true);
        break;
        case DPMOD_VALUE:
        c=strchr(dispval+1,' ');
        if (c) {
            drawBigText(display,0,15,dispval,WHITE,-1,false, c-dispval);
            drawBigText(display,128-18*strlen(c),15,c,WHITE,-1,true);
        }
        else drawBigText(display,0,15,dispval);
        break;
    }
    drawText9x15(-1,0,modenames[mainMode],1);  
    display.display();

    lmv_LiIon = mV_LiIon/10;
    lmv_Vcc = mV_Vcc/10;
    oldDPMOD = dpMOD;
    oldMainMode = mainMode;
    oldCalib = thisCalib;
    oldwifidisp = wifidisp;
    if (dpMOD == DPMOD_VALUE || dpMOD == DPMOD_CALIB) strcpy(olddispval, dispval);
    else olddispval[0] = 0;
    if (dpMOD == DPMOD_SELECT) oldSelShow = selShow;
    
}

