#ifndef SPEECHD_H
#define SPEECHD_H

#include <Arduino.h>
#ifdef ESP8266
#include <ESP8266WiFi.h>
#define ESPSTRING "esp8266"
#else
#include <WiFi.h>
#include <WiFiClient.h>
#define ESPSTRING "esp32"
#endif

class SpeechdClient
{

    public:
    SpeechdClient(const char *clientName = nullptr, const char *host=nullptr, int port=0, bool interactive=false);
    ~SpeechdClient(void);
    int setHost(const char *host, int port = 0);
    int setHost(uint32_t host, int port = 0);
    int setHost(String &host, int port = 0) { return setHost(host.c_str(), port); }
    int say(const char *str, int priority=0);
    int stop(void);
    int setModule(const char *module);
    int setLanguage(const char *language);
    int setVoice(const char *voice);
    int setRate(int rate);
    int loop(void);
    int status(void);
    void disconnect(void);
    enum : int {PRIO_UNCHANGED=0, PRIO_PROGRESS, PRIO_NOTIFICATION, PRIO_TEXT, PRIO_MESSAGE, PRIO_IMPORTANT, PRIO_MAX};
    static const int DEFAULT_PORT=6560;
    protected:
    int _say(const char *str, int priority);
    int writeln(const char * str, size_t len=0);
    int readline(char *buf = nullptr, int len=256);
    int response(int *msg = nullptr);
    int _setParam(const char *param, const char *value);
    int _setIParam(const char *param, int value);
    int _readNotify(void);
    int connect(void);

    char linebuf[256];
    WiFiClient client;
    bool is_connected;
    bool _interactive;
    char *_clientName;
    char *_host;
    int _port;
    char _module[32];
    char _language[10];
    char _voice[32];
    int _rate;
    int clientID;
    int _messageID;
    int _messageStatus;
};

#endif
